/* camdrv.c */
/* CAMAC device driver for CC/7700-PCI on Linux 2.6.x */
/* Created by Enomoto Sanshiro on 11 April 1999. */
/* Last updated by Enomoto Sanshiro on 27 October 2005. */


#if defined(USE_MODVERSIONS) && USE_MODVERSIONS
#  define MODVERSIONS
#  include <config/modversions.h>
#endif


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/ioctl.h>
#include <linux/poll.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include "camdrv.h"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Enomoto Sanshiro");
MODULE_DESCRIPTION("Camac Driver for TOYO-Corp. CC/7700-PCI Controller");

#define CC7700_VENDOR_ID 0x0001
#define CC7700_DEVICE_ID 0xcc77

#ifndef CAMDRV_MAJOR
#  define CAMDRV_MAJOR 0
#endif
static int camdrv_major = CAMDRV_MAJOR;
module_param(camdrv_major, int, S_IRUGO);

static const char* camdrv_name = "camdrv";
static unsigned long ioport;
static int irq;

static int open_count;
static unsigned crate_address = 0;
static wait_queue_head_t camdrv_wait_queue;
static int interrupt_count;
static int dev_id;

static int camdrv_open(struct inode* inode, struct file* filep);
static int camdrv_release(struct inode* inode, struct file* filep);
static int camdrv_ioctl(struct inode *inode, struct file *filep, unsigned int cmd, unsigned long arg);
static unsigned int camdrv_poll(struct file* filep, poll_table* table);
static irqreturn_t camdrv_interrupt(int irq, void* dev_id, struct pt_regs* regs);

static int set_crate(void);
static int initialize(void);
static int clear(void);
static int inhibit(void);
static int release_inhibit(void);
static int camac_action(unsigned naf, unsigned* data);
static int read_lam(unsigned* data);
static int enable_interrupt(void);
static int disable_interrupt(void);
static int wait_lam(unsigned time_out, unsigned* data);
static int detect_pci_device(unsigned vendor, unsigned device, unsigned long* ioport, unsigned int* irq);


static struct file_operations camdrv_fops = {
    .open = camdrv_open,
    .release = camdrv_release,
    .ioctl = camdrv_ioctl,
    .poll = camdrv_poll,
};


int __init init_module(void)
{
    int result;

    result = detect_pci_device(CC7700_VENDOR_ID, CC7700_DEVICE_ID, &ioport, &irq);
    if (result) {
        printk(KERN_WARNING "%s: unable to find CC/7700-PCI.\n", camdrv_name);
	return result;
    }
    printk(KERN_INFO "%s: CC/7700-PCI is detected at ioport 0x%04lx on irq %d.\n", camdrv_name, ioport, irq);

    result = register_chrdev(camdrv_major, camdrv_name, &camdrv_fops);
    if (result < 0) {
        printk(KERN_WARNING "%s: unable to get major %d\n", camdrv_name, camdrv_major);
        return result;
    }
    if (camdrv_major == 0) {
        camdrv_major = result;
    }
    init_waitqueue_head(&camdrv_wait_queue);

    printk(KERN_INFO "%s: at 0x%04lx on irq %d (major = %d).\n", camdrv_name, ioport, irq, camdrv_major);

    open_count = 0;

    return 0;
}

void __exit cleanup_module(void)
{
    unregister_chrdev(camdrv_major, camdrv_name);
    printk(KERN_INFO "%s: removed.\n", camdrv_name);
}

static int camdrv_open(struct inode* inode, struct file* filep)
{
    int result;

    if (open_count) {
        return -EBUSY;
    }

    result = request_irq(irq, camdrv_interrupt, SA_INTERRUPT | SA_SHIRQ, camdrv_name, &dev_id);
    if (result < 0) {
        printk(KERN_WARNING "%s: unable to request IRQ.\n", camdrv_name);
        return result;
    }

    try_module_get(THIS_MODULE);
    open_count++;

    return 0;
}

static int camdrv_release(struct inode* inode, struct file* filep)
{
    disable_interrupt();
    free_irq(irq, &dev_id);

    module_put(THIS_MODULE);
    open_count--;

    return 0;
}

static int camdrv_ioctl(struct inode* inode, struct file *filep, unsigned int cmd, unsigned long arg)
{
    unsigned parameter = 0, data = 0;
    unsigned *user_parameter_ptr, *user_data_ptr;
    int result = -EINVAL;

    user_parameter_ptr = (unsigned *) arg;
    user_data_ptr = (unsigned *) arg + 1;

    if (_IOC_TYPE(cmd) != CAMDRV_IOC_MAGIC) {
        return -EINVAL;
    }

    if (_IOC_DIR(cmd) & (_IOC_READ | _IOC_WRITE)) {
	if (get_user(parameter, user_parameter_ptr) < 0) {
	    return -EFAULT;
	}
	if (get_user(data, user_data_ptr) < 0) {
	    return -EFAULT;
	}
    }

    switch (cmd) {
      case CAMDRV_IOC_INITIALIZE:
        result = initialize();
        break;
      case CAMDRV_IOC_CLEAR:
        result = clear();
        break;
      case CAMDRV_IOC_INHIBIT:
        result = inhibit();
        break;
      case CAMDRV_IOC_RELEASE_INHIBIT:
        result = release_inhibit();
        break;
      case CAMDRV_IOC_ENABLE_INTERRUPT:
        result = enable_interrupt();
        break;
      case CAMDRV_IOC_DISABLE_INTERRUPT:
        result = disable_interrupt();
        break;
      case CAMDRV_IOC_CAMAC_ACTION:
        result = camac_action(parameter, &data);
        put_user(data, user_data_ptr);
        break;
      case CAMDRV_IOC_READ_LAM:
        result = read_lam(&data);
        put_user(data, user_data_ptr);
        break;
      case CAMDRV_IOC_WAIT_LAM:
        result = wait_lam(parameter, &data);
        put_user(data, user_data_ptr);
        break;
      case CAMDRV_IOC_SET_CRATE:
        crate_address = parameter;
        result = set_crate();
        break;
      default:
	result = -EINVAL;
    }

    return result;
}

static unsigned int camdrv_poll(struct file* filep, poll_table* table)
{
    unsigned int mask = 0;

    poll_wait(filep, &camdrv_wait_queue, table);

    if (interrupt_count > 0) {
	mask |= POLLIN | POLLRDNORM;
    }

    return mask;
}

static irqreturn_t camdrv_interrupt(int irq, void* dev_id, struct pt_regs* regs)
{
    unsigned lam_pattern;
    read_lam(&lam_pattern);

    if (lam_pattern == 0) {
        return IRQ_NONE;
    }

    disable_interrupt();
    interrupt_count++;
    wake_up_interruptible(&camdrv_wait_queue);

    return IRQ_HANDLED;
}


/* CAMAC service functions */

enum cc7700_regs {
    regCONTROL = 0x00,
    regSTATUS = 0x00,
    regCOMMAND = 0x04,
    regLAM = 0x04,
    regDATA = 0x08,
    regGO = 0x0C
};

enum cc7700_ctrlbits {
    ctrlCLEAR = 0x01,
    ctrlINITIALIZE = 0x02,
    ctrlINHIBIT = 0x04,
    ctrlENABLEINTERRUPT = 0x08,
    ctrlRESET = 0x40,
    ctrlLAMINTERNAL = 0x80
};

enum cc7700_statbits {
    statNOQ = 0x01,
    statNOX = 0x02,
    statINHIBIT = 0x04,
    statENABLEINTERRUPT = 0x08,
    statDONE = 0x10,
    statONLINE = 0x20,
    statLAMSUM = 0x40,
    statLAMINTERNAL = 0x80
};

#define CONTROL_UNTOUCHED_BIT_MASK (0x04 | 0x08 | 0x80)
#define STATUS_BIT_MASK 0x00ff

static int set_crate(void)
{
    unsigned command;
    static const unsigned naf = 0;

    command = (crate_address << 16) | naf;
    outl(command, ioport + regCOMMAND);

    return 0;
}

static int initialize(void)
{
    outl(ctrlINITIALIZE, ioport + regCONTROL);

    return 0;
}

static int clear(void)
{
    unsigned reg;

    reg = inl(ioport + regCONTROL) & CONTROL_UNTOUCHED_BIT_MASK;
    reg |= ctrlCLEAR;
    
    outl(reg, ioport + regCONTROL);

    return 0;
}

static int inhibit(void)
{
    unsigned reg;

    reg = inl(ioport + regCONTROL) & CONTROL_UNTOUCHED_BIT_MASK;
    reg |= ctrlINHIBIT;
    
    outl(reg, ioport + regCONTROL);

    return 0;
}

static int release_inhibit(void)
{
    unsigned reg;

    reg = inl(ioport + regCONTROL) & CONTROL_UNTOUCHED_BIT_MASK;
    reg &= ~ctrlINHIBIT;
    
    outl(reg, ioport + regCONTROL);

    return 0;
}

static int enable_interrupt(void)
{
    unsigned reg;

    interrupt_count = 0;
    reg = inl(ioport + regCONTROL) & CONTROL_UNTOUCHED_BIT_MASK;
    reg |= ctrlENABLEINTERRUPT;
    
    outl(reg, ioport + regCONTROL);

    return 0;
}

static int disable_interrupt(void)
{
    unsigned reg;

    reg = inl(ioport + regCONTROL) & CONTROL_UNTOUCHED_BIT_MASK;
    reg &= ~ctrlENABLEINTERRUPT;
    
    outl(reg, ioport + regCONTROL);

    return 0;
}

static int camac_action(unsigned naf, unsigned* data)
{
    unsigned command;
    unsigned data_mask = 0x00ffffff;
    unsigned long start_jiffies;
    int status;

    /* Hardware Problem of CC/7700 ??? */
    /* subtract 1 from station number (bit 9-13) */
    naf -= (1 << 9);
    
    command = (crate_address << 16) | naf;
    outl(command, ioport + regCOMMAND);
    outl(*data, ioport + regDATA);

    wmb();
    outl(0, ioport + regGO);

    start_jiffies = jiffies;
    while (! ((status = inl(ioport + regSTATUS)) & statDONE)) {
        schedule();
	if (jiffies - start_jiffies > HZ) {
	    *data = 0;
	    return -ETIMEDOUT;
	}
    }

    *data = inl(ioport + regDATA) & data_mask;

    return status & STATUS_BIT_MASK;
}

static int read_lam(unsigned* data)
{
    unsigned data_mask = 0x00ffffff;

    *data = inl(ioport + regLAM) & data_mask;
    return *data;
}

static int wait_lam(unsigned timeout, unsigned* data)
{
    long remaining_time;
    timeout *= HZ;

    remaining_time = wait_event_interruptible_timeout(
        camdrv_wait_queue, 
	(interrupt_count > 0),
	timeout
    );
    if (remaining_time == 0) {
        return -ETIMEDOUT;
    }
    if (interrupt_count == 0) {
        return -ERESTARTSYS;
    }

    interrupt_count = 0;

    return read_lam(data);
}

#ifndef CONFIG_PCI
#error "Kernel doesn't support PCI BIOS."
#endif

static int detect_pci_device(unsigned vendor, unsigned device, unsigned long* ioport, unsigned int* irq)
{
    struct pci_dev *dev;

    dev = NULL;
    dev = pci_find_device(vendor, device, dev);
    if (dev == NULL) {
        return -ENODEV;
    }

    *ioport = dev->resource[0].start & PCI_BASE_ADDRESS_IO_MASK;
    *irq = dev->irq;

    return 0;
}
